;
; ABSTRACO TABULEIRO
;
; Desenha um tabuleiro
; composto por n-x por n-y clulas quadradas,
; e devolve uma estrutura de dados com as caractersticas
; do tabuleiro desenhado.
; Permite alterar o contedo visual de qualquer clula,
; colocando nela um smbolo  escolha de um conjunto
; de vrios smbolos grficos.
;
; Sobre esta abstraco  possvel desenvolver aplicaes
; que tenham por base um tabuleiro (jogo de damas, 
; labirintos, ...), sem obrigar a conhecer o software 
; grfico que acompanha as vrias implementaes do Scheme.
;
; F. Nunes Ferreira
; fnf@fe.up.pt
; 4 de Novembro de 2000
;
;
;
;
;
; Modo de utilizao
;
; 1) Garantir que existe uma janela grfica, 
; suficientemente ampla para conter o tabuleiro.
;
; Ex:
; (janela 200 200 "tabuleiro")
;
; Nota: Para limpar a janela grfica
; (limpa)
;
;
; 2) Constructor de tabuleiro
; 
; tabuleiro: org-x org-y   (origem do vrtice sup-esq.)
;            lado (comprimento do lado das clulas)
;            num-cel-x num-cel-y (nmero de clulas)
;
;            As clulas so identificadas por dois 
;            ndices, situados entre 0 e num-cel-x - 1 
;            e 0 e num-cel-y - 1.
;
;
; Ex:
; (define t (tabuleiro 10 190 20 5 7))
; 
; Desenha um tabuleiro e devolve a lista
; (10 190 20 5 7)
;
;
; 3) Selectores tabuleiro
;
; (cel-x t) e (cel-y t) devolvem, respectivamente, 
; nmero de clulas em x e nmero de clulas em y
; do tabuleiro t.
;
;
; 4) Manipulao visual de clulas do tabuleiro
;
; 4.1)
; (celula tab i-x i-y simbolo)
;
; Altera o aspecto visual de uma clula,
; desenhando nela um smbolo  escolha.
;
; No tabuleiro tab, na clula i-x i-y,
; desenha um smbolo de acordo com:
; 
; simbolo    aco
; -------    -----
; +          desenha uma cruz +
; x          desenha uma crux x
; c          desenha crculo O
; p          desenha ponto .
; n          desenha seta para norte
; s          desenha seta para sul
; e          desenha seta para este
; o          desenha seta para oeste
; l          limpa clula
;
;
; 4.2)
; (celulas tab lista simb espera)
;
; Altera um conjunto de clulas, especificadas
; pelo parmetro lista, com o smbolo simb, a um ritmo
; condicionado por espera (quanto maior, mais lento...).
; O parmetro lista  uma lista de pares de indces, (cons i j).
;
; 4.3)
; (limpa-tab tab)
;
; Limpa o tabuleiro tab.
;
; 
;
;
; --------------------------
; --------------------------
;
; carrega software grfico disponibilizado no 
; mbito da disciplina IP1 (LEIC)
;

(require-library "swgr.scm" "user-feup")

;
; Cria um tabuleiro, visualiza-o e devolve uma
; lista com as caractersticas do tabuleiro criado.
; Os elementos desta lista so os argumentos 
; fornecidos na chamada.
;
(define tabuleiro
  (lambda (org-x org-y lado num-cel-x num-cel-y)
    
    (letrec ((linha-quad        ; desenha uma linha de ncx clulas 
              (lambda (ncx)
                (if (zero? ncx)
                    (move-rel (list (- (* num-cel-x
                                          lado))
                                    (- lado)))
                    (begin
                      (desenha-quadrado lado)
                      (move-rel (list lado 0))
                      (linha-quad (sub1 ncx))))))
             
             (tab              ; desenha tabuleiro de ncy linhas de clulas
              (lambda (ncy)
                (if (not (zero? ncy))
                    (begin
                      (linha-quad num-cel-x)
                      (tab (sub1 ncy)))))))
      ;
      (cor 0)
      (move (list org-x org-y))     ; caneta no canto superior-esquerdo
      (tab num-cel-y)
      (move (list org-x org-y))
      (list org-x org-y lado num-cel-x num-cel-y))))
; 
; Altera o aspecto visual de uma clula
;
(define celula
  (lambda (tab i j id)
    (cor 0)
    (let ((ox (car tab))
          (oy (cadr tab))
          (lado (caddr tab))
          (nx (cadddr tab))
          (ny (car (cddddr tab))))
      (let* ((lado/2 (/ lado 2))
             (lado/5 (/ lado 5))
             (lado2/5 (* 2 lado/5))
             (lado3/5 (* 3 lado/5))
             (lado4/5 (* 4 lado/5))
             (vx (+ ox (* i lado)))
             (vy (- oy (* j lado))))
        
        (let ((limpa         ; limpa clula
               (lambda()
                 (cor 15)
                 (move (list vx vy))
                 (pinta-quadrado lado)
                 (cor 0)
                 (desenha-quadrado lado)))
              
              (cruz+        ; desenha cruz +
               (lambda()
                 (move (list (+ vx lado/2)
                             (- vy lado/5)))
                 (desenha-rel (list (list 0 0)
                                    (list 0 (- (* 3 lado/5)))))
                 (move (list (+ vx lado/5)
                             (- vy lado/2)))
                 (desenha-rel (list (list 0 0)
                                    (list (+ (* 3 lado/5)) 0)))))
              
              (cruzx        ; desenha cruz x
               (lambda()
                 (move (list (+ vx lado/5)
                             (- vy lado/5)))
                 (desenha-rel (list (list 0 0)
                                    (list (+ (* 3 lado/5))
                                          (- (* 3 lado/5)))))
                 (move (list (+ vx lado/5)
                             (- vy lado4/5)))
                 (desenha-rel (list (list 0 0)
                                    (list (+ (* 3 lado/5))
                                          (+ (* 3 lado/5)))))))
              
              (circulo      ; desenha crculo  O
               (lambda()
                 (desenha-oval (list
                                (list (+ vx lado/5)
                                      (- vy lado/5))
                                (list (+ vx lado4/5)
                                      (- vy lado4/5))))))
              
              (ponto        ; desenha ponto .
               (lambda()
                 (pinta-oval (list
                              (list (+ vx lado2/5)
                                    (- vy lado2/5))
                              (list (+ vx lado3/5)
                                    (- vy lado3/5))))))
              
              (seta-norte   ; desenha seta "norte"
               (lambda()
                 (move (list (+ vx lado/5)
                             (- vy lado/2)))
                 (desenha-rel (list (list 0 0)
                                    (list (+ (* 1.5 lado/5))
                                          (+ (* 1.5 lado/5)))
                                    (list (+ (* 1.5 lado/5))
                                          (- (* 1.5 lado/5)))))
                 (move (list (+ vx lado/2)
                             (- vy lado4/5)))
                 (desenha-rel (list (list 0 0)
                                    (list 0
                                          (+ (* 3 lado/5)))))))
              
              (seta-sul   ; desenha seta "sul"
               (lambda()
                 (move (list (+ vx lado/5)
                             (- vy lado/2)))
                 (desenha-rel (list (list 0 0)
                                    (list (+ (* 1.5 lado/5))
                                          (- (* 1.5 lado/5)))
                                    (list (+ (* 1.5 lado/5))
                                          (+ (* 1.5 lado/5)))))
                 (move (list (+ vx lado/2)
                             (- vy lado4/5)))
                 (desenha-rel (list (list 0 0)
                                    (list 0
                                          (+ (* 3 lado/5)))))))
              
              (seta-este   ; desenha seta "este"
               (lambda()
                 (move (list (+ vx lado/2)
                             (- vy lado/5)))
                 (desenha-rel (list (list 0 0)
                                    (list (+ (* 1.5 lado/5))
                                          (- (* 1.5 lado/5)))
                                    (list (- (* 1.5 lado/5))
                                          (- (* 1.5 lado/5)))))
                 (move (list (+ vx lado/5)
                             (- vy lado/2)))
                 (desenha-rel (list (list 0 0)
                                    (list (+ (* 3 lado/5))
                                          0)))))
              
              (seta-oeste   ; desenha seta "oeste"
               (lambda()
                 (move (list (+ vx lado/2)
                             (- vy lado/5)))
                 (desenha-rel (list (list 0 0)
                                    (list (- (* 1.5 lado/5))
                                          (- (* 1.5 lado/5)))
                                    (list (+ (* 1.5 lado/5))
                                          (- (* 1.5 lado/5)))))
                 (move (list (+ vx lado/5)
                             (- vy lado/2)))
                 (desenha-rel (list (list 0 0)
                                    (list (+ (* 3 lado/5))
                                          0))))))
          (if (or (< i 0)
                  (>= i nx)
                  (< j 0)
                  (>= j ny))
              (display "erro: fora do tabuleiro!!")
              (cond ((equal? id 'l)
                     (limpa))
                    ((equal? id '+)
                     (cruz+))
                    ((equal? id 'x)
                     (cruzx))
                    ((equal? id 'c)
                     (circulo))
                    ((equal? id 'p)
                     (ponto))
                    ((equal? id 'n)
                     (seta-norte))
                    ((equal? id 's)
                     (seta-sul))
                    ((equal? id 'e)
                     (seta-este))
                    ((equal? id 'o)
                     (seta-oeste))
                    (else
                     (display "erro: comando desconhecido!")))))))))
;
;
; limpa tabuleiro
;
(define limpa-tab 
  (lambda (tab)
    (let ((dim-x (cel-x tab))
          (dim-y (cel-y tab)))
      (letrec ((aux
                (lambda (n)
                  (if (zero? n)
                      '()
                      (cons n
                            (aux (sub1 n))))))
               (converte-x-y 
                (lambda (dim lista-de-numeros)
                  (if (null? lista-de-numeros)
                      '()
                      (let ((primeiro (sub1 (car lista-de-numeros))))
                        (cons (cons (remainder primeiro dim)
                                    (quotient primeiro dim))
                              (converte-x-y dim (cdr lista-de-numeros))))))))
        
        (celulas tab 
                 (converte-x-y dim-x (aux (* dim-x dim-y))) 
                 'l 
                 0)))))
;
; Altera o aspecto visual de um conjunto de clulas, especificadas pelo
; parmetro lista, com o smbolo sim e a um ritmo dependente de espera.
;
(define celulas
  (lambda (tab lista simb espera)
    (letrec ((espera-aux
              (lambda (n)
                (if (not (zero? n))
                    (espera-aux (sub1 n))))))
      (if (not (null? lista))
          (begin 
            (celula tab (caar lista) (cdar lista) simb)
            (espera-aux espera)
            (celulas tab (cdr lista) simb espera))))))

;
; Selectores de dimenso do tabuleiro
;
; Nmero de clulas por linha
;
(define cel-x
  (lambda (tabul)
    (car (cdddr tabul))))
;
; Nmero de clulas por coluna
;
(define cel-y
  (lambda (tabul)
    (car (cddddr tabul))))

;
; ---------------------------------------------------
; procedimentos auxiliares
; ---------------------------------------------------
;
(define desenha-quadrado
  (lambda (lado)
    (desenha-rel (list (list 0 0)
                       (list lado 0)
                       (list 0 (- lado))
                       (list (- lado) 0)
                       (list 0 lado)))))
(define pinta-quadrado
  (lambda (lado)
    (pinta-rel (list (list 0 0)
                     (list lado 0)
                     (list 0 (- lado))
                     (list (- lado) 0)
                     (list 0 lado)))))